steps = [
    7, 8, 9, 10, 11, 12, 13, 14,
    16, 17, 19, 21, 23, 25, 28, 31,
    34, 37, 41, 45, 50, 55, 60, 66,
    73, 80, 88, 97, 107, 118, 130, 143,
    157, 173, 190, 209, 230, 253, 279, 307,
    337, 371, 408, 449, 494, 544, 598, 658,
    724, 796, 876, 963, 1060, 1166, 1282, 1411,
    1552, 1707, 1878, 2066, 2272, 2499, 2749, 3024,
    3327, 3660, 4026, 4428, 4871, 5358, 5894, 6484,
    7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
    15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794,
    32767
]

step_indices = [-1, -1, -1, -1, 2, 4, 6, 8]

def clamp(value, lower, upper):
    return lower if (value < lower) else upper if (value > upper) else value

def decode_sample(sample: int, state: tuple) -> tuple:
    predicted_sample = state[0]
    step_index = state[1]
    step = steps[step_index]

    diff = step >> 3
    if sample & 1: diff += step >> 2
    if sample & 2: diff += step >> 1
    if sample & 4: diff += step
    if sample & 8: diff = -diff
    
    predicted_sample = clamp(predicted_sample + diff, -32768, 32767)
    step_index = clamp(step_index + step_indices[sample & 7], 0, 88)

    return predicted_sample, step_index

def encode_sample(sample: int, state: tuple) -> tuple:
    predicted_sample = state[0]
    step_index = state[1]
    step = steps[step_index]

    sample_diff = sample - predicted_sample
    encoded_sample = 0

    if sample_diff < 0:
        sample_diff = -sample_diff
        encoded_sample = 8
    
    diff = step >> 3
    if sample_diff >= step:
        encoded_sample |= 4
        sample_diff -= step
        diff += step

    step >>= 1
    if sample_diff >= step:
        encoded_sample |= 2
        sample_diff -= step
        diff += step

    step >>= 1
    if sample_diff >= step:
        encoded_sample |= 1
        diff += step

    if encoded_sample & 8:
        diff = -diff
    
    predicted_sample = clamp(predicted_sample + diff, -32768, 32767)
    step_index = clamp(step_index + step_indices[encoded_sample & 7], 0, 88)

    return encoded_sample, (predicted_sample, step_index)

import wave, struct, time

def decode(samples: bytes) -> bytearray:
    samples = memoryview(samples).cast('B')
    decoded_samples = bytearray()
    state = (0, 0)

    for sample in samples:
        state = decode_sample(sample & 0xF, state)
        decoded_samples += struct.pack('<h', state[0])
        state = decode_sample((sample >> 4) & 0xF, state)
        decoded_samples += struct.pack('<h', state[0])
    return decoded_samples

def encode(samples: bytes) -> bytearray:
    samples = memoryview(samples).cast('h')
    encoded_samples = bytearray()
    state = (0, 0)
    low_sample = 0
    high_sample = False

    for sample in samples:
        sample, state = encode_sample(sample, state)

        if high_sample:
            encoded_samples.append(sample << 4 | low_sample)
        else:
            low_sample = sample

        high_sample = not high_sample
    return encoded_samples


import json, glob
from pathlib import Path
from argparse import ArgumentParser
from struct import pack, unpack, calcsize
import xml.etree.ElementTree as ET

header_fmt = '< 2I' #magic_number, version
header_size = calcsize(header_fmt)

element_fmt = '< I 2i I' #name_offset, next_element_offset, sub_element_offset, attr_count
element_size = calcsize(element_fmt)

attr_fmt = '< 2I' #name_offset, value_offset
attr_size = calcsize(attr_fmt)

def read_string(xmlb_file, offset: int) -> str:
    last_pos = xmlb_file.tell()
    string = b''

    xmlb_file.seek(offset)

    while char := xmlb_file.read(1):
        if char == b'\x00':
            break
        string += char
        
    xmlb_file.seek(last_pos)
    return string.decode('cp1252')

def read_element(xmlb_file) -> tuple[ET.Element, int]:
    name_offset, next_element_offset, sub_element_offset, attr_count = unpack(element_fmt, xmlb_file.read(element_size))
    element = ET.Element(read_string(xmlb_file, name_offset))
    offset = -1
    
    for i in range(attr_count):
        name_offset, value_offset = unpack(attr_fmt, xmlb_file.read(attr_size))
        element.set(read_string(xmlb_file, name_offset), read_string(xmlb_file, value_offset))

    if sub_element_offset != -1:
        xmlb_file.seek(sub_element_offset)
        sub_element, offset = read_element(xmlb_file)
        element.append(sub_element)

    while offset != -1:
        xmlb_file.seek(offset)
        next_element, offset = read_element(xmlb_file)
        element.append(next_element)

    return (element, next_element_offset)

def read_xmlb(xmlb_path: Path) -> ET.Element:
    with xmlb_path.open(mode='rb') as xmlb_file:
        magic_number, version = unpack(header_fmt, xmlb_file.read(header_size))

        if magic_number != 0x11B1 or version != 1:
            raise ValueError('Invalid magic number')
        
        root_element, next_element_offset = read_element(xmlb_file)
        return root_element

#http://effbot.org/zone/element-lib.htm#prettyprint
def indent(elem, level=0):
    i = "\n" + level*"    "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "    "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent(elem, level+1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i

#https://stackoverflow.com/a/29520802/15020406
class FakeDict(dict):
    def __init__(self, items):
        self['something'] = 'something'
        self._items = items
    def items(self):
        return self._items

def str2value(string: str):
    value = string

    try:
        value = int(value)

        if len(str(value)) < len(string):
            value = string
    except ValueError:
        try:
            value = float(value)
        except ValueError:
            if value == 'true':
                value = True
            elif value == 'false':
                value = False
    return value

def to_json_element(element: ET.Element) -> tuple[str, FakeDict]:
    elements = []

    for name, value in element.items():
        elements.append((name, str2value(value)))

    for sub_element in element:
        elements.append(to_json_element(sub_element))

    return (element.tag, FakeDict(elements))

def decompile(xmlb_path: Path, output_path: Path, has_indent: bool):
    root_element = read_xmlb(xmlb_path)

    if output_path.suffix == '.xml':
        if has_indent:
            indent(root_element)
        ET.ElementTree(root_element).write(output_path, encoding='utf-8')
    elif output_path.suffix == '.json':
        with output_path.open(mode='w', encoding='utf-8') as json_file:
            json.dump(FakeDict([to_json_element(root_element)]), json_file, indent=4 if (has_indent) else None, ensure_ascii=False)
    else:
        raise ValueError(f'Output file extension {output_path.suffix} is not supported. Supported: .xml, .json')

string_offsets = {}
next_string_offset = header_size

def get_offset(key: str) -> int:
    global next_string_offset

    if key not in string_offsets:
        string_offsets[key] = next_string_offset
        next_string_offset += len(key) + 1
    
    return string_offsets[key]

def write_element(xmlb_file, element: ET.Element, has_next: bool):
    attr_count = len(element.attrib)
    sub_element_count = len(element)
    sub_element_offset = xmlb_file.tell() + element_size + attr_count * attr_size if (sub_element_count > 0) else -1

    if has_next:
        next_element_offset = xmlb_file.tell()

        for sub_element in element.iter():
            next_element_offset += element_size + len(sub_element.attrib) * attr_size
    else:
        next_element_offset = -1

    xmlb_file.write(pack(element_fmt, get_offset(element.tag), next_element_offset, sub_element_offset, attr_count))

    for name, value in element.items():
        xmlb_file.write(pack(attr_fmt, get_offset(name), get_offset(value)))

    for sub_element_index, sub_element in enumerate(element):
        write_element(xmlb_file, sub_element, sub_element_index < sub_element_count - 1)

def write_xmlb(root_element: ET.Element, output_path: Path):
    global next_string_offset

    with output_path.open(mode='wb') as xmlb_file:
        for element in root_element.iter():
            next_string_offset += element_size + len(element.attrib) * attr_size

        xmlb_file.write(pack(header_fmt, 0x11B1, 1))
        write_element(xmlb_file, root_element, False)

        for key in string_offsets:
            xmlb_file.write(key.encode('cp1252'))
            xmlb_file.write(b'\x00')

        string_offsets.clear()
        next_string_offset = header_size

def parse_json_object_pairs(pairs):
    return pairs

def value2str(value) -> str:
    if isinstance(value, str):
        return value

    string = str(value)
    return string.lower() if (isinstance(value, bool)) else string

def from_json_element(element: tuple) -> ET.Element:
    tag, sub_elements = element
    xml_element = ET.Element(tag)

    for element in sub_elements:
        tag, value = element
        
        if isinstance(value, list):
            xml_element.append(from_json_element(element))
        else:
            xml_element.set(tag, value2str(value))

    return xml_element

def compile(input_path: Path, output_path: Path):
    if input_path.suffix == '.xml':
        write_xmlb(ET.parse(input_path).getroot(), output_path)
    elif input_path.suffix == '.json':
        with input_path.open(mode='r', encoding='utf-8') as json_file:
            data = json.load(json_file, object_pairs_hook=parse_json_object_pairs)
            root_amount = len(data)

            if root_amount != 1:
                raise ValueError(f'Found {root_amount} root elements. Required 1')

            write_xmlb(from_json_element(data[0]), output_path)
    else:
        raise ValueError(f'Input file extension {input_path.suffix} is not supported. Supported: .xml, .json')

def main():
    parser = ArgumentParser()
    parser.add_argument('-d', '--decompile', action='store_true', help='decompile input XMLB file to XML/JSON file')
    parser.add_argument('--no_indent', action='store_true', help='disable indent in decompiled XML/JSON file')
    parser.add_argument('input', help='input file (supports glob)')
    parser.add_argument('output', help='output file (wildcards will be replaced by input file name)')
    args = parser.parse_args()
    input_files = glob.glob(glob.escape(args.input), recursive=True)

    if not input_files:
        raise ValueError('No files found')

    for input_file in input_files:
        input_file = Path(input_file)
        output_file = Path(args.output.replace('*', input_file.stem))
        output_file.parent.mkdir(parents=True, exist_ok=True)

        if args.decompile:
            decompile(input_file, output_file, not args.no_indent)
        else:
            compile(input_file, output_file)

if __name__ == '__main__':
    main()



import json, glob
from pathlib import Path
from argparse import ArgumentParser
from struct import Struct
from io import BytesIO
import xml.etree.ElementTree as ET

FBFileHeader = Struct(
    '128s' # file path
    '64s' # file type
    'I' # file size
)

XML_F = ['xml', 'eng', 'fre', 'ger', 'ita', 'spa', 'rus', 'pol']
XML_Formats = []
for f in XML_F:
    XML_Formats.append('.' + f)
    XML_Formats.append('.' + f + 'b')

Known_Formats = {
    'actors.igb': 'actorskin',
    'anim.igb': 'actoranimdb',
    'textures.igb': 'texture',
    'conversations.xmlb': 'xml',
    'data.xmlb': 'xml',
    'weapons.xmlb': 'xml',
    'entities.xmlb': 'xml',
    'talents.xmlb': 'xml_talents',
    'powerstyles.xmlb': 'fightstyle',
    'fightstyles.xmlb': 'fightstyle',
    'shared_nodes.xmlb': 'fightstyle_xml',
    'effects.xmlb': 'effect',
    'maps.xmlb': 'zonexml',
    'motionpaths.igb': 'motionpath',
    'shared_powerups.xmlb': 'shared_powerups',
    '.xmlb': 'xml_resident',
    '.igb': 'model',
    '.py': 'script',
    '.chrb': 'characters',
    '.chr': 'characters',
    '.navb': 'nav',
    '.nav': 'nav',
    '.boyb': 'boy',
    '.boy': 'boy',
    '.pkgb': 'pkg',
    '.pkg': 'pkg',
    '.zam': 'zam',
    '.shd': 'shadow',
    '.sdfb': 'sdf',
    '.sdf': 'sdf'
}

def decompile(fb_path: Path, output_path: Path):
    fb_data = fb_path.read_bytes()

    with BytesIO(fb_data) as fb_file:
        entries = ET.Element('packagedef')

        while (file_header := fb_file.read(FBFileHeader.size)):
            file_path, file_type, file_size = FBFileHeader.unpack(file_header)
            file_path = file_path.decode().split('\x00', 1)[0]
            file_type = file_type.decode().split('\x00', 1)[0]
            file_data = fb_file.read(file_size)

            child = ET.SubElement(entries, file_type)
            #inner text: child.text = str(file_path)
            child.set('filename', file_path) #attribute
            file_path = output_path.parent / output_path.stem / file_path
            file_path.parent.mkdir(parents=True, exist_ok=True)
            file_path.write_bytes(file_data)

        ET.indent(entries, ' ' * 4)
        ET.ElementTree(entries).write(output_path, encoding='utf-8')

def compile(xml_path: Path, output_path: Path):
    data = ET.parse(xml_path)

    with BytesIO() as fb_data:
        for e in data.findall("./*"):
            file_path = e.attrib['filename']
            file_type = e.tag
            real_file_path = xml_path.parent / xml_path.stem / file_path
            file_data = real_file_path.read_bytes()
            fb_file_header = FBFileHeader.pack(file_path.encode(), file_type.encode(), len(file_data))

            fb_data.write(fb_file_header)
            fb_data.write(file_data)

        output_path.write_bytes(fb_data.getbuffer())

def rebuild(xml_path: Path, output_path: Path):
    data = ET.parse(xml_path)
    input_folder = xml_path.parent / xml_path.stem

    with BytesIO() as fb_data:
        for file_type in ('combat_is', 'bigconvmap'):
            e = data.find('./' + file_type)
            if e is not None:
                file_path = e.attrib['filename']
                fb_file_header = FBFileHeader.pack(file_path.encode(), file_type.encode(), 0)
                fb_data.write(fb_file_header)

        for p in input_folder.rglob("*.*"):
            file_type = ''
            file_path = str(p.relative_to(input_folder)).replace('\\', '/')
            for e in data.findall("./*"):
                if e.attrib['filename'] == file_path:
                    file_type = e.tag
            if file_type == '':
                folders = file_path.split('/')
                sf = folders[1] if len(folders) > 1 else ''
                dp = sf.rsplit('.', maxsplit=1)[0]
                folder = sf if folders[0] == 'data' and '.' not in sf else 'anim' if folders[0] == 'actors' and not all(d in '0123456789' for d in p.stem) else dp if dp == 'shared_powerups' else dp[0:12] if dp[0:12] == 'shared_nodes' else folders[0]
                e = p.suffix.lower()
                if e in XML_Formats: e = '.xmlb'
                type_string = folder + e
                file_type = Known_Formats[type_string] if type_string in Known_Formats else Known_Formats[e] if e in Known_Formats else 'unknown'
            file_data = p.read_bytes()
            fb_file_header = FBFileHeader.pack(file_path.encode(), file_type.encode(), len(file_data))

            fb_data.write(fb_file_header)
            fb_data.write(file_data)

        output_path.write_bytes(fb_data.getbuffer())

def main():
    parser = ArgumentParser()
    parser.add_argument('-d', '--decompile', action='store_true', help='decompile input FB file to JSON file')
    parser.add_argument('-r', '--rebuild', action='store_true', help='compile to FB file, including all files that exist in the corresponding directory')
    parser.add_argument('input', help='input file (supports glob)')
    parser.add_argument('output', help='output file (wildcards will be replaced by input file name)')
    args = parser.parse_args()
    input_files = glob.glob(args.input, recursive=True)

    if not input_files:
        raise ValueError('No files found')

    for input_file in input_files:
        input_file = Path(input_file)
        output_file = Path(args.output.replace('*', input_file.stem))
        output_file.parent.mkdir(parents=True, exist_ok=True)

        if args.decompile:
            decompile(input_file, output_file)
        elif args.rebuild:
            rebuild(input_file, output_file)
        else:
            compile(input_file, output_file)

if __name__ == '__main__':
    main()